{-# OPTIONS -Wwarn #-}
{-# LANGUAGE OverlappingInstances #-}
{-# LANGUAGE NoPatternSynonyms #-}
{-# LANGUAGE CPP #-}
#if __GLASGOW_HASKELL__ <= 708
{-# LANGUAGE OverlappingInstances #-}
#endif
{-# LANGUAGE FlexibleInstances #-}
{-# OPTIONS_GHC -fno-warn-incomplete-patterns #-}

-- | Pretty-printer for Tog.
--   Generated by the BNF converter.

module Tog.Raw.Print where

import qualified Tog.Raw.Abs
import Data.Char

-- | The top-level printing method.

printTree :: Print a => a -> String
printTree = render . prt 0

type Doc = [ShowS] -> [ShowS]

doc :: ShowS -> Doc
doc = (:)

render :: Doc -> String
render d = rend 0 (map ($ "") $ d []) "" where
  rend i ss = case ss of
    "["      :ts -> showChar '[' . rend i ts
    "("      :ts -> showChar '(' . rend i ts
    "{"      :ts -> showChar '{' . new (i+1) . rend (i+1) ts
    "}" : ";":ts -> new (i-1) . space "}" . showChar ';' . new (i-1) . rend (i-1) ts
    "}"      :ts -> new (i-1) . showChar '}' . new (i-1) . rend (i-1) ts
    ";"      :ts -> showChar ';' . new i . rend i ts
    t  : ts@(p:_) | closingOrPunctuation p -> showString t . rend i ts
    t        :ts -> space t . rend i ts
    _            -> id
  new i   = showChar '\n' . replicateS (2*i) (showChar ' ') . dropWhile isSpace
  space t = showString t . (\s -> if null s then "" else ' ':s)

  closingOrPunctuation :: String -> Bool
  closingOrPunctuation [c] = c `elem` closerOrPunct
  closingOrPunctuation _   = False

  closerOrPunct :: String
  closerOrPunct = ")],;"

parenth :: Doc -> Doc
parenth ss = doc (showChar '(') . ss . doc (showChar ')')

concatS :: [ShowS] -> ShowS
concatS = foldr (.) id

concatD :: [Doc] -> Doc
concatD = foldr (.) id

replicateS :: Int -> ShowS -> ShowS
replicateS n f = concatS (replicate n f)

-- | The printer class does the job.

class Print a where
  prt :: Int -> a -> Doc
  prtList :: Int -> [a] -> Doc
  prtList i = concatD . map (prt i)

instance {-# OVERLAPPABLE #-} Print a => Print [a] where
  prt = prtList

instance Print Char where
  prt _ s = doc (showChar '\'' . mkEsc '\'' s . showChar '\'')
  prtList _ s = doc (showChar '"' . concatS (map (mkEsc '"') s) . showChar '"')

mkEsc :: Char -> Char -> ShowS
mkEsc q s = case s of
  _ | s == q -> showChar '\\' . showChar s
  '\\'-> showString "\\\\"
  '\n' -> showString "\\n"
  '\t' -> showString "\\t"
  _ -> showChar s

prPrec :: Int -> Int -> Doc -> Doc
prPrec i j = if j < i then parenth else id

instance Print Integer where
  prt _ x = doc (shows x)

instance Print Double where
  prt _ x = doc (shows x)

instance Print Tog.Raw.Abs.Name where
  prt _ (Tog.Raw.Abs.Name (_,i)) = doc (showString i)
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print Tog.Raw.Abs.Empty where
  prt _ (Tog.Raw.Abs.Empty (_,i)) = doc (showString i)

instance Print Tog.Raw.Abs.Module where
  prt i e = case e of
    Tog.Raw.Abs.Module name params decls -> prPrec i 0 (concatD [doc (showString "module"), prt 0 name, prt 0 params, doc (showString "where"), doc (showString "{"), prt 0 decls, doc (showString "}")])

instance Print Tog.Raw.Abs.QName where
  prt i e = case e of
    Tog.Raw.Abs.Qual qname name -> prPrec i 0 (concatD [prt 0 qname, doc (showString "."), prt 0 name])
    Tog.Raw.Abs.NotQual name -> prPrec i 0 (concatD [prt 0 name])

instance Print [Tog.Raw.Abs.Decl] where
  prt = prtList

instance Print Tog.Raw.Abs.Decl where
  prt i e = case e of
    Tog.Raw.Abs.TypeSig typesig -> prPrec i 0 (concatD [prt 0 typesig])
    Tog.Raw.Abs.FunDef name patterns fundefbody -> prPrec i 0 (concatD [prt 0 name, prt 0 patterns, prt 0 fundefbody])
    Tog.Raw.Abs.Data name params databody -> prPrec i 0 (concatD [doc (showString "data"), prt 0 name, prt 0 params, prt 0 databody])
    Tog.Raw.Abs.Record name params recordbody -> prPrec i 0 (concatD [doc (showString "record"), prt 0 name, prt 0 params, prt 0 recordbody])
    Tog.Raw.Abs.OpenImport import_ openspecs -> prPrec i 0 (concatD [doc (showString "open"), doc (showString "import"), prt 0 import_, prt 0 openspecs])
    Tog.Raw.Abs.OpenApp import_ openspecs -> prPrec i 0 (concatD [doc (showString "open"), prt 0 import_, prt 0 openspecs])
    Tog.Raw.Abs.Import import_ -> prPrec i 0 (concatD [doc (showString "import"), prt 0 import_])
    Tog.Raw.Abs.ModuleApp name import_ -> prPrec i 0 (concatD [doc (showString "module"), prt 0 name, doc (showString "="), prt 0 import_])
    Tog.Raw.Abs.Postulate typesigs -> prPrec i 0 (concatD [doc (showString "postulate"), doc (showString "{"), prt 0 typesigs, doc (showString "}")])
    Tog.Raw.Abs.Mutual decls -> prPrec i 0 (concatD [doc (showString "mutual"), doc (showString "{"), prt 0 decls, doc (showString "}")])
    Tog.Raw.Abs.Module_ module_ -> prPrec i 0 (concatD [prt 0 module_])
    Tog.Raw.Abs.Pragma -> prPrec i 0 (concatD [doc (showString "{-@AGDA-}")])
  prtList _ [] = concatD []
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, doc (showString ";"), prt 0 xs]

instance Print [Tog.Raw.Abs.OpenSpec] where
  prt = prtList

instance Print Tog.Raw.Abs.OpenSpec where
  prt i e = case e of
    Tog.Raw.Abs.Public -> prPrec i 0 (concatD [doc (showString "public")])
  prtList _ [] = concatD []
  prtList _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Tog.Raw.Abs.TypeSig] where
  prt = prtList

instance Print Tog.Raw.Abs.Import where
  prt i e = case e of
    Tog.Raw.Abs.ImportNoArgs qname -> prPrec i 0 (concatD [prt 0 qname])
    Tog.Raw.Abs.ImportArgs qname args -> prPrec i 0 (concatD [prt 0 qname, prt 0 args])

instance Print Tog.Raw.Abs.TypeSig where
  prt i e = case e of
    Tog.Raw.Abs.Sig name expr -> prPrec i 0 (concatD [prt 0 name, doc (showString ":"), prt 0 expr])
  prtList _ [] = concatD []
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, doc (showString ";"), prt 0 xs]

instance Print Tog.Raw.Abs.Where where
  prt i e = case e of
    Tog.Raw.Abs.Where decls -> prPrec i 0 (concatD [doc (showString "where"), doc (showString "{"), prt 0 decls, doc (showString "}")])
    Tog.Raw.Abs.NoWhere -> prPrec i 0 (concatD [])

instance Print Tog.Raw.Abs.Params where
  prt i e = case e of
    Tog.Raw.Abs.NoParams -> prPrec i 0 (concatD [])
    Tog.Raw.Abs.ParamDecl bindings -> prPrec i 0 (concatD [prt 0 bindings])
    Tog.Raw.Abs.ParamDef hiddennames -> prPrec i 0 (concatD [prt 0 hiddennames])

instance Print Tog.Raw.Abs.HiddenName where
  prt i e = case e of
    Tog.Raw.Abs.NotHidden name -> prPrec i 0 (concatD [prt 0 name])
    Tog.Raw.Abs.Hidden name -> prPrec i 0 (concatD [doc (showString "{"), prt 0 name, doc (showString "}")])
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Tog.Raw.Abs.HiddenName] where
  prt = prtList

instance Print Tog.Raw.Abs.DataBody where
  prt i e = case e of
    Tog.Raw.Abs.DataDecl name -> prPrec i 0 (concatD [doc (showString ":"), prt 0 name])
    Tog.Raw.Abs.DataDef constrs -> prPrec i 0 (concatD [doc (showString "where"), prt 0 constrs])
    Tog.Raw.Abs.DataDeclDef name constrs -> prPrec i 0 (concatD [doc (showString ":"), prt 0 name, doc (showString "where"), prt 0 constrs])

instance Print Tog.Raw.Abs.Constrs where
  prt i e = case e of
    Tog.Raw.Abs.Constrs constrs -> prPrec i 0 (concatD [doc (showString "{"), prt 0 constrs, doc (showString "}")])
    Tog.Raw.Abs.NoConstrs -> prPrec i 0 (concatD [doc (showString "{"), doc (showString "{-@EMPTY-}"), doc (showString "}")])

instance Print Tog.Raw.Abs.RecordBody where
  prt i e = case e of
    Tog.Raw.Abs.RecordDecl name -> prPrec i 0 (concatD [doc (showString ":"), prt 0 name])
    Tog.Raw.Abs.RecordDef name fields -> prPrec i 0 (concatD [doc (showString "where"), doc (showString "{"), doc (showString "constructor"), prt 0 name, prt 0 fields, doc (showString "}")])
    Tog.Raw.Abs.RecordDeclDef name1 name2 fields -> prPrec i 0 (concatD [doc (showString ":"), prt 0 name1, doc (showString "where"), doc (showString "{"), doc (showString "constructor"), prt 0 name2, prt 0 fields, doc (showString "}")])

instance Print Tog.Raw.Abs.Fields where
  prt i e = case e of
    Tog.Raw.Abs.NoFields -> prPrec i 0 (concatD [])
    Tog.Raw.Abs.Fields constrs -> prPrec i 0 (concatD [doc (showString ";"), doc (showString "field"), doc (showString "{"), prt 0 constrs, doc (showString "}")])

instance Print Tog.Raw.Abs.Constr where
  prt i e = case e of
    Tog.Raw.Abs.Constr name expr -> prPrec i 0 (concatD [prt 0 name, doc (showString ":"), prt 0 expr])
  prtList _ [] = concatD []
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, doc (showString ";"), prt 0 xs]

instance Print [Tog.Raw.Abs.Constr] where
  prt = prtList

instance Print Tog.Raw.Abs.FunDefBody where
  prt i e = case e of
    Tog.Raw.Abs.FunDefNoBody -> prPrec i 0 (concatD [])
    Tog.Raw.Abs.FunDefBody expr where_ -> prPrec i 0 (concatD [doc (showString "="), prt 0 expr, prt 0 where_])

instance Print Tog.Raw.Abs.Telescope where
  prt i e = case e of
    Tog.Raw.Abs.Tel bindings -> prPrec i 0 (concatD [prt 0 bindings])

instance Print Tog.Raw.Abs.Binding where
  prt i e = case e of
    Tog.Raw.Abs.Bind args expr -> prPrec i 0 (concatD [doc (showString "("), prt 0 args, doc (showString ":"), prt 0 expr, doc (showString ")")])
    Tog.Raw.Abs.HBind args expr -> prPrec i 0 (concatD [doc (showString "{"), prt 0 args, doc (showString ":"), prt 0 expr, doc (showString "}")])
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Tog.Raw.Abs.Binding] where
  prt = prtList

instance Print Tog.Raw.Abs.Expr where
  prt i e = case e of
    Tog.Raw.Abs.Lam names expr -> prPrec i 0 (concatD [doc (showString "\\"), prt 0 names, doc (showString "->"), prt 0 expr])
    Tog.Raw.Abs.Pi telescope expr -> prPrec i 0 (concatD [prt 0 telescope, doc (showString "->"), prt 0 expr])
    Tog.Raw.Abs.Fun expr1 expr2 -> prPrec i 0 (concatD [prt 1 expr1, doc (showString "->"), prt 0 expr2])
    Tog.Raw.Abs.Eq expr1 expr2 -> prPrec i 1 (concatD [prt 2 expr1, doc (showString "=="), prt 2 expr2])
    Tog.Raw.Abs.App args -> prPrec i 2 (concatD [prt 0 args])
    Tog.Raw.Abs.Id qname -> prPrec i 3 (concatD [prt 0 qname])

instance Print Tog.Raw.Abs.Arg where
  prt i e = case e of
    Tog.Raw.Abs.HArg expr -> prPrec i 0 (concatD [doc (showString "{"), prt 0 expr, doc (showString "}")])
    Tog.Raw.Abs.Arg expr -> prPrec i 0 (concatD [prt 3 expr])
  prtList _ [x] = concatD [prt 0 x]
  prtList _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Tog.Raw.Abs.Arg] where
  prt = prtList

instance Print Tog.Raw.Abs.Pattern where
  prt i e = case e of
    Tog.Raw.Abs.EmptyP empty -> prPrec i 0 (concatD [prt 0 empty])
    Tog.Raw.Abs.ConP qname patterns -> prPrec i 0 (concatD [doc (showString "("), prt 0 qname, prt 0 patterns, doc (showString ")")])
    Tog.Raw.Abs.IdP qname -> prPrec i 0 (concatD [prt 0 qname])
    Tog.Raw.Abs.HideP pattern -> prPrec i 0 (concatD [doc (showString "{"), prt 0 pattern, doc (showString "}")])
  prtList _ [] = concatD []
  prtList _ (x:xs) = concatD [prt 0 x, prt 0 xs]

instance Print [Tog.Raw.Abs.Pattern] where
  prt = prtList

instance Print [Tog.Raw.Abs.Name] where
  prt = prtList

